from typing import Optional

from flowchem.components.properties import PassiveComponent


class PassiveMixer(PassiveComponent):
    """
    A generic mixer (essentially an alias of PassiveComponent).
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)



from .tube import Tube
from .channel import Channel
from .cross_mixer import CrossMixer
from .t_mixer import TMixer
from .y_mixer import YMixer
from .pump import Pump
from .vessel import Vessel
from .vessel_chemicals import VesselChemicals



"""
HA Elite11 tests
Run with python -m pytest ./tests -m HApump and updates pump com port and address in pump below
"""
import asyncio
import math

import pytest

from flowchem.components.devices.Harvard_Apparatus.HA_elite11 import (
    Elite11InfuseWithdraw,
    PumpStatus
)
from flowchem.units import flowchem_ureg


async def move_infuse(pump):
    await pump.set_syringe_diameter(10)
    await pump.set_infusion_rate(1)
    await pump.set_target_volume(1)
    await pump.infuse_run()


@pytest.fixture(scope="session")
def event_loop(request):
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
async def pump():
    """ Change to match your hardware ;) """
    pump = Elite11InfuseWithdraw.from_config(port="COM11", syringe_volume=5, diameter=20)
    await pump.initialize()
    return pump


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_version(pump: Elite11InfuseWithdraw):
    assert "11 ELITE" in await pump.version()


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_status_idle(pump: Elite11InfuseWithdraw):
    await pump.stop()
    assert await pump.get_status() is PumpStatus.IDLE


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_status_infusing(pump: Elite11InfuseWithdraw):
    await move_infuse(pump)
    assert await pump.get_status() is PumpStatus.INFUSING
    await pump.stop()


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_status_withdrawing(pump: Elite11InfuseWithdraw):
    await pump.set_syringe_diameter(10)
    await pump.set_withdraw_rate(1)
    await pump.withdraw_run()
    assert await pump.get_status() is PumpStatus.WITHDRAWING
    await pump.stop()


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_is_moving(pump: Elite11InfuseWithdraw):
    assert await pump.is_moving() is False
    await move_infuse(pump)
    assert await pump.is_moving() is True
    await pump.stop()


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_syringe_volume(pump: Elite11InfuseWithdraw):
    await pump.set_syringe_volume(10)
    assert await pump.get_syringe_volume() == "10 ml"
    await pump.set_syringe_volume(math.pi)
    vol = flowchem_ureg.Quantity(await pump.get_syringe_volume()).magnitude
    assert math.isclose(vol, math.pi, abs_tol=10e-4)
    await pump.set_syringe_volume(3e-05)
    vol = flowchem_ureg.Quantity(await pump.get_syringe_volume()).magnitude
    assert math.isclose(vol, 3e-5)
    await pump.set_syringe_volume(50)  # Leave a sensible value otherwise other tests will fail!


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_infusion_rate(pump: Elite11InfuseWithdraw):
    await pump.set_syringe_diameter(10)
    await pump.set_infusion_rate(5)
    assert await pump.get_infusion_rate()
    with pytest.warns(UserWarning):
        await pump.set_infusion_rate(121)
    rate = flowchem_ureg.Quantity(await pump.get_infusion_rate()).magnitude
    assert math.isclose(rate, 12.49, rel_tol=0.01)
    with pytest.warns(UserWarning):
        await pump.set_infusion_rate(0)
    rate = flowchem_ureg.Quantity(await pump.get_infusion_rate()).magnitude
    assert math.isclose(rate, 1e-05, abs_tol=1e-5)
    await pump.set_infusion_rate(math.pi)
    rate = flowchem_ureg.Quantity(await pump.get_infusion_rate()).magnitude
    assert math.isclose(rate, math.pi, abs_tol=0.001)


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_get_infused_volume(pump: Elite11InfuseWithdraw):
    await pump.clear_volumes()
    assert await pump.get_infused_volume() == "0 ul"
    await pump.set_syringe_diameter(30)
    await pump.set_infusion_rate(5)
    await pump.set_target_volume(0.05)
    await pump.infuse_run()
    await asyncio.sleep(2)
    vol = flowchem_ureg.Quantity(await pump.get_infused_volume()).to("ml").magnitude
    assert math.isclose(vol, 0.05, abs_tol=1e-4)


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_get_withdrawn_volume(pump: Elite11InfuseWithdraw):
    await pump.clear_volumes()
    await pump.set_withdraw_rate(10)
    await pump.set_target_volume(0.1)
    await pump.withdraw_run()
    await asyncio.sleep(1)
    vol = flowchem_ureg.Quantity(await pump.get_withdrawn_volume()).to("ml").magnitude
    assert math.isclose(vol, 0.1, abs_tol=1e-4)


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_force(pump: Elite11InfuseWithdraw):
    await pump.set_force(10)
    assert await pump.get_force() == 10
    await pump.set_force(50.2)
    assert await pump.get_force() == 50
    assert await pump.get_force() == 50


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_diameter(pump: Elite11InfuseWithdraw):
    await pump.set_syringe_diameter(10)
    assert await pump.get_syringe_diameter() == "10.0000 mm"

    with pytest.warns(UserWarning):
        await pump.set_syringe_diameter(34)

    with pytest.warns(UserWarning):
        await pump.set_syringe_diameter(0.01)

    await pump.set_syringe_diameter(math.pi)
    dia = flowchem_ureg.Quantity(await pump.get_syringe_diameter()).magnitude
    math.isclose(dia, math.pi)


@pytest.mark.HApump
@pytest.mark.asyncio
async def test_target_volume(pump: Elite11InfuseWithdraw):
    await pump.set_syringe_volume(10)
    await pump.set_target_volume(math.pi)
    vol = flowchem_ureg.Quantity(await pump.get_target_volume()).magnitude
    assert math.isclose(vol, math.pi, abs_tol=10e-4)
    await pump.set_target_volume(1e-04)
    vol = flowchem_ureg.Quantity(await pump.get_target_volume()).magnitude
    assert math.isclose(vol, 1e-4, abs_tol=10e-4)



""" Test FlowIR, needs actual connection to the device :( """
import asyncio
import datetime
import sys

import pytest

from flowchem.components.devices.MettlerToledo.iCIR_common import IRSpectrum
from flowchem import FlowIR


def check_pytest_asyncio_installed():
    """ Utility function for pytest plugin """
    import os
    from importlib import util
    if not util.find_spec("pytest_asyncio"):
        print("You need to install pytest-asyncio first!", file=sys.stderr)
        sys.exit(os.EX_SOFTWARE)


@pytest.fixture()
async def flowir():
    """ Return local FlowIR object """
    return FlowIR(FlowIR.iC_OPCUA_DEFAULT_SERVER_ADDRESS.replace("localhost", "BSMC-YMEF002121"))


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_connected(flowir):
    async with flowir as spectrometer:
        assert await spectrometer.is_iCIR_connected()


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_probe_info(flowir):
    async with flowir as spectrometer:
        info = await spectrometer.probe_info()
        assert all(field in info for field in ("spectrometer", "spectrometer_SN", "probe_SN", "detector"))


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_probe_status(flowir):
    async with flowir as spectrometer:
        status = await spectrometer.probe_status()
        assert status == "Not running"


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_is_running(flowir):
    async with flowir as spectrometer:
        # Check idle
        assert await spectrometer.is_running() is False

        # Make busy
        template_name = "15_sec_integration.iCIRTemplate"
        await spectrometer.start_experiment(template=template_name)

        # Check busy
        assert await spectrometer.is_running() is True

        # Restore idle
        await spectrometer.stop_experiment()
        await spectrometer.wait_until_idle()


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_spectrum_acquisition(flowir):
    async with flowir as spectrometer:
        template_name = "15_sec_integration.iCIRTemplate"
        await spectrometer.start_experiment(template=template_name)

        # Wait for a spectrum
        spectrum = await spectrometer.last_spectrum_treated()
        while spectrum.empty:
            spectrum = await spectrometer.last_spectrum_treated()

        assert isinstance(spectrum, IRSpectrum)

        spectra_count = await spectrometer.sample_count()

        while await spectrometer.sample_count() == spectra_count:
            await asyncio.sleep(1)

        # sample_count() increments as expected
        assert await spectrometer.sample_count() == spectra_count + 1

        # Check date
        date = await spectrometer.last_sample_time()
        assert isinstance(date, datetime.datetime)
        assert date.date() == datetime.date.today()

        # Restore idle
        await spectrometer.stop_experiment()
        await spectrometer.wait_until_idle()


@pytest.mark.asyncio
@pytest.mark.FlowIR
async def test_spectra(flowir):
    # This implies the previous test run successfully, thus last spectrum is now
    async with flowir as spectrometer:
        last_raw = await spectrometer.last_spectrum_raw()
        assert isinstance(last_raw, IRSpectrum)

        last_bg = await spectrometer.last_spectrum_background()
        assert isinstance(last_bg, IRSpectrum)



"""
This module is used to control Hamilton ML600 syringe pump via the protocol1/RNO+.
"""

from __future__ import annotations

from flowchem.components.stdlib import Pump
from loguru import logger
import string
import time
import warnings
from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, Set, TYPE_CHECKING

if TYPE_CHECKING:
    import pint

import aioserial

from flowchem.exceptions import InvalidConfiguration, DeviceError
from flowchem.units import flowchem_ureg


@dataclass
class Protocol1CommandTemplate:
    """Class representing a pump command and its expected reply, but without target pump number"""

    command: str
    optional_parameter: str = ""
    execute_command: bool = True

    def to_pump(
        self, address: int, command_value: str = "", argument_value: str = ""
    ) -> Protocol1Command:
        """Returns a Protocol11Command by adding to the template pump address and command arguments"""
        return Protocol1Command(
            target_pump_num=address,
            command=self.command,
            optional_parameter=self.optional_parameter,
            command_value=command_value,
            argument_value=argument_value,
            execute_command=self.execute_command,
        )


@dataclass
class Protocol1Command(Protocol1CommandTemplate):
    """Class representing a pump command and its expected reply"""

    PUMP_ADDRESS = {
        pump_num: address
        for (pump_num, address) in enumerate(string.ascii_lowercase[:16], start=1)
    }
    # i.e. PUMP_ADDRESS = {1: 'a', 2: 'b', 3: 'c', 4: 'd', ..., 16: 'p'}
    # Note ':' is used for broadcast within the daisy chain.

    target_pump_num: int = 1
    command_value: Optional[str] = None
    argument_value: Optional[str] = None

    def compile(self) -> bytes:
        """Create actual command byte by prepending pump address to command and appending executing command."""
        assert self.target_pump_num in range(1, 17)
        if not self.command_value:
            self.command_value = ""

        compiled_command = (
            f"{self.PUMP_ADDRESS[self.target_pump_num]}"
            f"{self.command}{self.command_value}"
        )

        if self.argument_value:
            compiled_command += f"{self.optional_parameter}{self.argument_value}"
        # Add execution flag at the end
        if self.execute_command is True:
            compiled_command += "R"

        return (compiled_command + "\r").encode("ascii")


class HamiltonPumpIO:
    """Setup with serial parameters, low level IO"""

    ACKNOWLEDGE = chr(6)
    NEGATIVE_ACKNOWLEDGE = chr(21)
    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 9600,
        "parity": aioserial.PARITY_EVEN,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.SEVENBITS,
    }

    def __init__(self, aio_port: aioserial.Serial):
        """
        Initialize communication on the serial port where the pumps are located and initialize them
        Args:
            aio_port: aioserial.Serial() object
        """
        self._serial = aio_port

        # These will be set by `HamiltonPumpIO.initialize()`
        self._initialized = False
        self.num_pump_connected: Optional[int] = None

    @classmethod
    def from_config(cls, config):
        """Create HamiltonPumpIO from config."""
        # Merge default settings, including serial, with provided ones.
        configuration = dict(HamiltonPumpIO.DEFAULT_CONFIG, **config)

        try:
            serial_object = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the pump on the port <{configuration.get('port')}>"
            ) from e

        return cls(serial_object)

    async def initialize(self, hw_initialization: bool = True):
        """
        Ensure connection with pump + initialize

        Args:
            hw_initialization: Whether each pump has to be initialized. Note that this might be undesired!
        """
        # This has to be run after each power cycle to assign addresses to pumps
        self.num_pump_connected = await self._assign_pump_address()
        if hw_initialization:
            await self._hw_init()
        self._initialized = True

    async def _assign_pump_address(self) -> int:
        """
        To be run on init, auto assign addresses to pumps based on their position in the daisy chain.
        A custom command syntax with no addresses is used here so read and write has been rewritten
        """
        try:
            await self._write_async("1a\r".encode("ascii"))
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e

        reply = await self._read_reply_async()
        if reply and reply[:1] == "1":
            # reply[1:2] should be the address of the last pump. However, this does not work reliably.
            # So here we enumerate the pumps explicitly instead
            last_pump = 0
            for pump_num, address in Protocol1Command.PUMP_ADDRESS.items():
                await self._write_async(f"{address}UR\r".encode("ascii"))
                if "NV01" in await self._read_reply_async():
                    last_pump = pump_num
                else:
                    break
            logger.debug(f"Found {last_pump} pumps on {self._serial.port}!")
            return int(last_pump)
        else:
            raise InvalidConfiguration(f"No pump found on {self._serial.port}")

    async def _hw_init(self):
        """Send to all pumps the HW initialization command (i.e. homing)"""
        await self._write_async(b":XR\r")  # Broadcast: initialize + execute
        # Note: no need to consume reply here because there is none (since we are using broadcast)

    async def _write_async(self, command: bytes):
        """Writes a command to the pump"""
        if not self._initialized:
            raise DeviceError(
                "Pump not initialized!\n"
                "Have you called `initialize()` after object creation?"
            )
        await self._serial.write_async(command)
        logger.debug(f"Command {repr(command)} sent!")

    async def _read_reply_async(self) -> str:
        """Reads the pump reply from serial communication"""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string}")
        return reply_string.decode("ascii")

    def parse_response(self, response: str) -> str:
        """Split a received line in its components: success, reply"""
        status = response[:1]
        assert status in (
            HamiltonPumpIO.ACKNOWLEDGE,
            HamiltonPumpIO.NEGATIVE_ACKNOWLEDGE,
        ), "Invalid status reply!"

        if status == HamiltonPumpIO.ACKNOWLEDGE:
            logger.debug("Positive acknowledge received")
        else:
            logger.warning("Negative acknowledge received")
            warnings.warn(
                "Negative acknowledge reply received from pump: check command validity!"
            )

        return response[1:].rstrip()

    def reset_buffer(self):
        """Reset input buffer before reading from serial. In theory not necessary if all replies are consumed..."""
        self._serial.reset_input_buffer()

    async def write_and_read_reply_async(self, command: Protocol1Command) -> str:
        """Main HamiltonPumpIO method.
        Sends a command to the pump, read the replies and returns it, optionally parsed"""
        self.reset_buffer()
        await self._write_async(command.compile())
        response = await self._read_reply_async()

        if not response:
            raise InvalidConfiguration(
                f"No response received from pump, check pump address! "
                f"(Currently set to {command.target_pump_num})"
            )

        return self.parse_response(response)

    @property
    def name(self) -> str:
        """This is used to provide a nice-looking default name to pumps based on their serial connection."""
        try:
            return self._serial.name
        except AttributeError:
            return ""


class ML600(Pump):
    """ ML600 implementation according to docs. Tested on 61501-01 (single syringe).

    From docs:
    To determine the volume dispensed per step the total syringe volume is divided by
    48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
    length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
    example to dispense 9 mL from a 10 mL syringe you would determine the number of
    steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
    """

    # This class variable is used for daisy chains (i.e. multiple pumps on the same serial connection). Details below.
    _io_instances: Set[HamiltonPumpIO] = set()
    # The mutable object (a set) as class variable creates a shared state across all the instances.
    # When several pumps are daisy-chained on the same serial port, they need to all access the same Serial object,
    # because access to the serial port is exclusive by definition (also locking there ensure thread safe operations).
    # FYI it is a borg idiom https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s23.html

    class ValvePositionName(IntEnum):
        """Maps valve position to the corresponding number"""

        POSITION_1 = 1
        # POSITION_2 = 2
        POSITION_3 = 3
        INPUT = 9  # 9 is default inlet, i.e. 1
        OUTPUT = 10  # 10 is default outlet, i.e. 3
        WASH = 11  # 11 is default wash, i.e. undefined

    # Only Hamilton syringes are compatible w/ the ML600, and they come on a limited set of sizes. (Values in ml)
    VALID_SYRINGE_VOLUME = {
        0.01,
        0.025,
        0.05,
        0.1,
        0.25,
        0.5,
        1.0,
        2.5,
        5.0,
        10.0,
        25.0,
        50.0,
    }

    def __init__(
        self,
        pump_io: HamiltonPumpIO,
        syringe_volume: str,
        address: int = 1,
        name: Optional[str] = None,
    ):
        """
        Default constructor, needs an HamiltonPumpIO object. See from_config() class method for config-based init.

        Args:
            pump_io: An HamiltonPumpIO w/ serial connection to the daisy chain w/ target pump.
            syringe_volume: Volume of the syringe used, either a Quantity or number in ml.
            address: number of pump in array, 1 for first one, auto-assigned on init based on position.
            name: 'cause naming stuff is important.
        """
        super().__init__(name)
        # HamiltonPumpIO
        self.pump_io = pump_io
        ML600._io_instances.add(self.pump_io)  # See above for details.

        # Pump address is the pump sequence number if in chain. Count starts at 1, default.
        self.address = int(address)

        # The pump name is used for logs and error messages.
        self.name = f"Pump {self.pump_io.name}:{address}" if name is None else name

        # Syringe pumps only perform linear movement, and the volume displaced is function of the syringe loaded.
        try:
            self.syringe_volume = flowchem_ureg(syringe_volume)
        except AttributeError as e:
            raise InvalidConfiguration(f"{self.__class__.__name__}:{self.name} "
                                       f"Syringe volume must be a string parsable as pint.Quantity!\n"
                                       f"It is now a {type(syringe_volume)}: {syringe_volume} ") from e

        if self.syringe_volume.m_as("ml") not in ML600.VALID_SYRINGE_VOLUME:
            raise InvalidConfiguration(
                f"The specified syringe volume ({syringe_volume}) does not seem to be valid!\n"
                f"The volume in ml has to be one of {ML600.VALID_SYRINGE_VOLUME}"
            )

        self._steps_per_ml = flowchem_ureg.Quantity(
            f"{48000 / self.syringe_volume} step/ml"
        )
        self._offset_steps = 100  # Steps added to each absolute move command, to decrease wear and tear at volume = 0
        self._max_vol = (
            (48000 - self._offset_steps) * flowchem_ureg.step / self._steps_per_ml
        )

    @classmethod
    def from_config(cls, **config):
        """This class method is used to create instances via config file by the server for HTTP interface."""
        # Many pump can be present on the same serial port with different addresses.
        # This shared list of HamiltonPumpIO objects allow shared state in a borg-inspired way, avoiding singletons
        # This is only relevant to programmatic instantiation, i.e. when from_config() is called per each pump from a
        # config file, as it is the case in the HTTP server.
        # HamiltonPump_IO() manually instantiated are not accounted for.
        pumpio = None
        for obj in ML600._io_instances:
            # noinspection PyProtectedMember
            if obj._serial.port == config.get("port"):
                pumpio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if pumpio is None:
            # Remove ML600-specific keys to only have HamiltonPumpIO's kwargs
            config_for_pumpio = {
                k: v
                for k, v in config.items()
                if k not in ("syringe_volume", "address", "name")
            }
            pumpio = HamiltonPumpIO.from_config(config_for_pumpio)

        return cls(
            pumpio,
            syringe_volume=config.get("syringe_volume"),
            address=config.get("address"),
            name=config.get("name"),
        )

    async def initialize(self, hw_init=False, init_speed: str = "200 sec / stroke"):
        """Must be called after init before anything else."""
        # Test connectivity by querying the pump's firmware version
        fw_cmd = Protocol1CommandTemplate(command="U").to_pump(self.address)
        firmware_version = await self.pump_io.write_and_read_reply_async(fw_cmd)
        logger.info(
            f"Connected to Hamilton ML600 {self.name} - FW version: {firmware_version}!"
        )

        if hw_init:
            await self.initialize_pump(speed=init_speed)

    async def send_command_and_read_reply(
        self,
        command_template: Protocol1CommandTemplate,
        command_value="",
        argument_value="",
    ) -> str:
        """Sends a command based on its template by adding pump address and parameters, returns reply"""
        return await self.pump_io.write_and_read_reply_async(
            command_template.to_pump(self.address, command_value, argument_value)
        )

    def _validate_speed(self, speed_value: Optional[str]) -> str:
        """Given a speed (seconds/stroke) returns a valid value for it, and a warning if out of bounds."""

        # Validated speeds are used as command argument, with empty string being the default for None
        if speed_value is None:
            return ""

        speed = flowchem_ureg(speed_value)

        # Alert if out of bounds but don't raise exceptions, according to general philosophy.
        # Target flow rate too high
        if speed < flowchem_ureg("2 sec/stroke"):
            speed = flowchem_ureg("2 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A bigger syringe is needed?"
            )

        # Target flow rate too low
        if speed > flowchem_ureg("3692 sec/stroke"):
            speed = flowchem_ureg("3692 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A smaller syringe is needed?"
            )

        return str(round(speed.m_as("sec / stroke")))

    async def initialize_pump(self, speed: Optional[str] = None):
        """
        Initialize both syringe and valve
        speed: 2-3692 in seconds/stroke
        """
        init_cmd = Protocol1CommandTemplate(command="X", optional_parameter="S")
        return await self.send_command_and_read_reply(
            init_cmd, argument_value=self._validate_speed(speed)
        )

    async def initialize_valve(self):
        """Initialize valve only"""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="LX")
        )

    async def initialize_syringe(self, speed: Optional[str] = None):
        """
        Initialize syringe only
        speed: 2-3692 in seconds/stroke
        """
        init_syringe_cmd = Protocol1CommandTemplate(
            command="X1", optional_parameter="S"
        )
        return await self.send_command_and_read_reply(
            init_syringe_cmd, argument_value=self._validate_speed(speed)
        )

    def flowrate_to_seconds_per_stroke(self, flowrate: str):
        """
        Convert flow rates to steps per seconds

        To determine the volume dispensed per step the total syringe volume is divided by
        48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
        length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
        example to dispense 9 mL from a 10 mL syringe you would determine the number of
        steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
        """
        flowrate = flowchem_ureg(flowrate)
        flowrate_in_steps_sec = flowrate * self._steps_per_ml
        seconds_per_stroke = (1 / flowrate_in_steps_sec).to("second/stroke")

        return self._validate_speed(str(seconds_per_stroke))

    def _seconds_per_stroke_to_flowrate(
        self, second_per_stroke: pint.Quantity
    ) -> float:
        """The inverse of flowrate_to_seconds_per_stroke(). Only internal use."""
        flowrate = 1 / (second_per_stroke * self._steps_per_ml)
        return flowrate.to("ml/min")

    def _volume_to_step_position(self, volume_w_units: str) -> int:
        """Converts a volume to a step position."""
        # noinspection PyArgumentEqualDefault
        volume = flowchem_ureg(volume_w_units)
        steps = volume * self._steps_per_ml
        return round(steps.m_as("steps")) + self._offset_steps

    async def _to_step_position(self, position: int, speed: str = ""):
        """Absolute move to step position."""
        abs_move_cmd = Protocol1CommandTemplate(command="M", optional_parameter="S")
        return await self.send_command_and_read_reply(
            abs_move_cmd, str(position), self._validate_speed(speed)
        )

    async def get_current_volume(self) -> str:
        """Return current syringe position in ml."""
        syringe_pos = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="YQP")
        )
        current_steps = (int(syringe_pos) - self._offset_steps) * flowchem_ureg.step
        return str(current_steps / self._steps_per_ml)

    async def to_volume(self, target_volume: str, speed: str = ""):
        """Absolute move to volume provided."""
        await self._to_step_position(
            self._volume_to_step_position(target_volume), speed
        )
        logger.debug(f"Pump {self.name} set to volume {target_volume} at speed {speed}")

    async def pause(self):
        """Pause any running command."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="K", execute_command=False)
        )

    async def resume(self):
        """Resume any paused command."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="$", execute_command=False)
        )

    async def stop(self):
        """Stops and abort any running command."""
        await self.pause()
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="V", execute_command=False)
        )

    async def wait_until_idle(self):
        """Returns when no more commands are present in the pump buffer."""
        logger.debug(f"ML600 pump {self.name} wait until idle...")
        while self.is_busy:
            time.sleep(0.1)
        logger.debug(f"...ML600 pump {self.name} idle now!")

    async def version(self) -> str:
        """Returns the current firmware version reported by the pump."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="U")
        )

    async def is_idle(self) -> bool:
        """Checks if the pump is idle (actually check if the last command has ended)."""
        return (
            await self.send_command_and_read_reply(
                Protocol1CommandTemplate(command="F")
            )
            == "Y"
        )

    async def is_busy(self) -> bool:
        """Pump is not idle."""
        return not await self.is_idle()

    async def get_valve_position(self) -> ValvePositionName:
        """Represent the position of the valve: getter returns Enum, setter needs Enum."""
        valve_pos = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="LQP")
        )
        return ML600.ValvePositionName(int(valve_pos))

    async def set_valve_position(
        self, target_position: ValvePositionName, wait_for_movement_end: bool = True
    ):
        """Set valve position. wait_for_movement_end is defaulted to True as it is a common mistake not to wait..."""
        valve_by_name_cw = Protocol1CommandTemplate(command="LP0")
        await self.send_command_and_read_reply(
            valve_by_name_cw, command_value=str(int(target_position))
        )
        logger.debug(f"{self.name} valve position set to {target_position.name}")
        if wait_for_movement_end:
            await self.wait_until_idle()

    async def get_return_steps(self) -> int:
        """Return steps' getter. Applied to the end of a downward syringe movement to removes mechanical slack."""
        steps = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="YQN")
        )
        return int(steps)

    async def set_return_steps(self, target_steps: int):
        """Return steps' setter. Applied to the end of a downward syringe movement to removes mechanical slack."""
        set_return_steps_cmd = Protocol1CommandTemplate(command="YSN")
        await self.send_command_and_read_reply(
            set_return_steps_cmd, command_value=str(int(target_steps))
        )

    async def pickup(
        self,
        volume: str,
        from_valve: ValvePositionName,
        flowrate: str = "1 ml/min",
        wait: bool = False,
    ):
        """Get volume from valve specified at given flowrate."""
        cur_vol = flowchem_ureg(await self.get_current_volume())
        if (cur_vol + volume) > self._max_vol:
            warnings.warn(
                f"Cannot withdraw {volume} given the current syringe position {cur_vol} and a "
                f"syringe volume of {self.syringe_volume}"
            )
            return

        # Valve to position specified
        await self.set_valve_position(from_valve)
        # Move up to target volume
        await self.to_volume(
            str(cur_vol + volume),
            speed=self.flowrate_to_seconds_per_stroke(flowrate),
        )

        if wait:
            await self.wait_until_idle()

    async def deliver(
        self,
        volume: str,
        to_valve: ValvePositionName,
        flowrate: str,
        wait: bool = False,
    ):
        """Delivers volume to valve specified at given flow rate."""
        cur_vol = flowchem_ureg(await self.get_current_volume())
        if volume > cur_vol:
            warnings.warn(
                f"Cannot deliver {volume} given the current syringe position {cur_vol}!"
            )
            return

        # Valve to position specified
        await self.set_valve_position(to_valve)
        # Move up to target volume
        await self.to_volume(
            str(cur_vol - volume),
            speed=self.flowrate_to_seconds_per_stroke(flowrate),
        )

        if wait:
            await self.wait_until_idle()

    async def transfer(
        self,
        volume: str,
        from_valve: ValvePositionName,
        to_valve: ValvePositionName,
        flowrate_in: str = "1 ml/min",
        flowrate_out: str = "1 ml/min",
        wait: bool = False,
    ):
        """Move liquid from place to place."""
        await self.pickup(volume, from_valve, flowrate_in, wait=True)
        await self.deliver(volume, to_valve, flowrate_out, wait=wait)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/firmware-version", self.version, methods=["GET"])
        router.add_api_route("/initialize/pump", self.initialize_pump, methods=["PUT"])
        router.add_api_route(
            "/initialize/valve", self.initialize_valve, methods=["PUT"]
        )
        router.add_api_route(
            "/initialize/syringe", self.initialize_syringe, methods=["PUT"]
        )
        router.add_api_route("/pause", self.pause, methods=["PUT"])
        router.add_api_route("/resume", self.resume, methods=["PUT"])
        router.add_api_route("/resume", self.resume, methods=["PUT"])
        router.add_api_route("/stop", self.stop, methods=["PUT"])
        router.add_api_route("/version", self.stop, methods=["PUT"])
        router.add_api_route("/is-idle", self.is_idle, methods=["GET"])
        router.add_api_route("/is-busy", self.is_busy, methods=["GET"])
        router.add_api_route(
            "/valve/position", self.get_valve_position, methods=["GET"]
        )
        router.add_api_route(
            "/valve/position", self.set_valve_position, methods=["PUT"]
        )
        router.add_api_route(
            "/syringe/volume", self.get_current_volume, methods=["GET"]
        )
        router.add_api_route("/syringe/volume", self.to_volume, methods=["PUT"])
        router.add_api_route(
            "/syringe/return-steps", self.get_return_steps, methods=["GET"]
        )
        router.add_api_route(
            "/syringe/return-steps", self.set_return_steps, methods=["PUT"]
        )
        router.add_api_route("/pickup", self.pickup, methods=["PUT"])
        router.add_api_route("/deliver", self.deliver, methods=["PUT"])
        # router.add_api_route("/transfer", self.transfer, methods=["PUT"])  # Might go in timeout

        return router


# class TwoPumpAssembly(Thread):
#     """
#     Thread to control two pumps and have them generating a continuous flow.
#     Note that the pumps should not be accessed directly when used in a TwoPumpAssembly!
#
#     Notes: this needs to start a thread owned by the instance to control the pumps.
#     The async version of this being possibly simpler w/ tasks and callback :)
#     """
#
#     def __init__(
#         self, pump1: ML600, pump2: ML600, target_flowrate: str, init_seconds: int = 10
#     ):
#         super(TwoPumpAssembly, self).__init__()
#         self._p1 = pump1
#         self._p2 = pump2
#         self.daemon = True
#         self.cancelled = threading.Event()
#         self._flowrate = ensure_quantity(target_flowrate, "ml/min")
#         logger = logging.getLogger(__name__).getChild("TwoPumpAssembly")
#         # How many seconds per stroke for first filling? application dependent, as fast as possible, but not too much.
#         self.init_secs = init_seconds
#
#         # While in principle possible, using syringes of different volumes is discouraged, hence...
#         assert (
#             pump1.syringe_volume == pump2.syringe_volume
#         ), "Syringes w/ equal volume are needed for continuous flow!"
#
#     async def initialize(self):
#         """ Initialize multi-pump """
#         await self._p1.initialize()
#         await self._p2.initialize()
#
#     @property
#     def flowrate(self):
#         """ Returns/sets flowrate. """
#         return self._flowrate
#
#     @flowrate.setter
#     def flowrate(self, target_flowrate):
#         if target_flowrate == 0:
#             warnings.warn(
#                 "Cannot set flowrate to 0! Pump stopped instead, restart previous flowrate with resume!"
#             )
#             self.cancel()
#         else:
#             self._flowrate = target_flowrate
#
#         # This will stop current movement, make wait_for_both_pumps() return and move on w/ updated speed
#         self._p1.stop()
#         self._p2.stop()
#
#     async def wait_for_both_pumps(self):
#         """ Custom waiting method to wait a shorter time than normal (for better sync) """
#         while await self._p1.is_busy() or await self._p2.is_busy():
#             await asyncio.sleep(0.01)  # 10ms sounds reasonable to me
#         logger.debug("Both pumps are ready!")
#
#     def _speed(self):
#         speed = self._p1.flowrate_to_seconds_per_stroke(self._flowrate)
#         logger.debug(f"Speed calculated as {speed}")
#         return speed
#
#     async def execute_stroke(
#         self, pump_full: ML600, pump_empty: ML600, speed_s_per_stroke: int
#     ):
#         """ Perform a cycle (1 syringe stroke) in the continuous-operation mode. See also run(). """
#         # Logic is a bit complex here to ensure pause-less pumping
#         # This needs the pump that withdraws to move faster than the pumping one. no way around.
#
#         # First start pumping with the full syringe already prepared
#         pump_full.to_volume(0, speed=speed_s_per_stroke)
#         logger.debug("Pumping...")
#         # Then start refilling the empty one
#         pump_empty.set_valve_position(pump_empty.ValvePositionName.INPUT)
#         # And do that fast so that we finish refill before the pumping is over
#         pump_empty.to_volume(pump_empty.syringe_volume, speed=speed_s_per_stroke - 5)
#         pump_empty.wait_until_idle()
#         # This allows us to set the right pump position on the pump that was empty (not full and ready for next cycle)
#         pump_empty.set_valve_position(pump_empty.ValvePositionName.OUTPUT)
#         pump_full.wait_until_idle()
#
#     def run(self):
#         """Overloaded Thread.run, runs the update
#         method once per every 10 milliseconds."""
#         # First initialize with init_secs speed...
#         self._p1.to_volume(self._p1.syringe_volume, speed=self.init_secs)
#         self._p1.wait_until_idle()
#         self._p1.valve_position = self._p1.ValvePositionName.OUTPUT
#         logger.info("Pumps initialized for continuous pumping!")
#
#         while True:
#             while not self.cancelled.is_set():
#                 self.execute_stroke(
#                     self._p1, self._p2, speed_s_per_stroke=self._speed()
#                 )
#                 self.execute_stroke(
#                     self._p2, self._p1, speed_s_per_stroke=self._speed()
#                 )
#
#     def cancel(self):
#         """ Cancel continuous-pumping assembly """
#         self.cancelled.set()
#         self._p1.stop()
#         self._p2.stop()
#
#     def resume(self):
#         """ Resume continuous-pumping assembly """
#         self.cancelled.clear()
#
#     def stop_and_return_solution_to_container(self):
#         """ Let´s not waste our precious stock solutions ;) """
#         self.cancel()
#         logger.info(
#             "Returning the solution currently loaded in the syringes back to the inlet.\n"
#             "Make sure the container is not removed yet!"
#         )
#         # Valve to input
#         self._p1.valve_position = self._p1.ValvePositionName.INPUT
#         self._p2.valve_position = self._p2.ValvePositionName.INPUT
#         self.wait_for_both_pumps()
#         # Volume to 0 with the init speed (supposedly safe for this application)
#         self._p1.to_volume(0, speed=self.init_secs)
#         self._p2.to_volume(0, speed=self.init_secs)
#         self.wait_for_both_pumps()
#         logger.info("Pump flushing completed!")


if __name__ == "__main__":
    import asyncio

    conf = {
        "port": "COM12",
        "address": 1,
        "name": "test1",
        "syringe_volume": 5,
    }
    pump1 = ML600.from_config(**conf)
    asyncio.run(pump1.initialize_pump())



from flowchem.components.properties import Component
from flowchem.units import flowchem_ureg


class Channel(Component):
    """
    A reaction channel.

    Arguments:
    - `length`: The length of the channel as a str.
    - `volume`: The channel volume as a str.
    - `material`: The material around the channel.

    """

    _id_counter = 0

    def __init__(self, length: str, volume: str, material: str, name: str = None):
        """
        See the `Tube` attributes for a description of the arguments.
        """
        self.length = flowchem_ureg.parse_expression(length)
        self.volume = flowchem_ureg.parse_expression(volume)

        self.material = material

        super().__init__(name)

    def __repr__(self):
        return f"Channel of length {self.length} and volume {self.volume}"



from .base_component import Component
from .active_component import ActiveComponent
from .passive_component import PassiveComponent
from .mapped_component import MultiportComponentMixin
from .sensor import Sensor
from .valve import Valve
from .injection_valve import InjectionValve
from .passivemixer import PassiveMixer
from .tempcontrol import TempControl



""" LTF reactors """
from typing import Optional

from flowchem.components.stdlib import Channel, YMixer
from flowchem.assemblies import Assembly


class LTF_HTM_ST_3_1(Assembly):
    """ An LTF HTM ST 3 1 reactor. """

    def _validate(self, dry_run):
        return True

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self.port = {"INLET_1", "INLET_2", "QUENCHER", "OUTLET"}

        inlet1 = Channel(name="INLET_1", length="10 mm", volume="8 ul", material="glass")
        inlet2 = Channel(name="INLET_2", length="10 mm", volume="8 ul", material="glass")
        mixer_inlet = YMixer()
        reactor1 = Channel(name="REACTOR", length="60 mm", volume="58 ul", material="glass")
        quencher = Channel(name="QUENCHER", length="15 mm", volume="10 ul", material="glass")
        mixer_quencher = YMixer()
        reactor2 = Channel(name="REACTOR2", length="40 mm", volume="46 ul", material="glass")
        outlet = Channel(name="OUTLET", length="10 mm", volume="28 ul", material="glass")

        self.nodes = [inlet1, inlet2, mixer_inlet, reactor1, quencher, mixer_quencher, reactor2, outlet]
        self.edges = [
            (inlet1, mixer_inlet),
            (inlet2, mixer_inlet),
            (mixer_inlet, reactor1),
            (reactor1, mixer_quencher),
            (quencher, mixer_quencher),
            (mixer_quencher, reactor2),
            (reactor2, outlet)
        ]



""" Represent a generic injection valve. """
from abc import ABC
from typing import Optional

from flowchem.components.properties import MultiportComponentMixin, ActiveComponent


class InjectionValve(MultiportComponentMixin, ActiveComponent, ABC):
    """
    A generic injection valve, i.e. a valve with positions 'inject' and 'load'.
    """

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        # For injection valves, the positions are 'load' and 'inject'
        self.position = {"inject", "load"}
        self.setting = "load"

        # Call Valve init
        super().__init__(name=name)

        # Ensure base state is loading.
        self._base_state = {"setting": "load"}

        # TODO add injection loop volume



from .dummy import Dummy
from .broken_dummy_component import BrokenDummyComponent
from .broken_dummy_sensor import BrokenDummySensor
from .dummy_pump import DummyPump
from .dummy_valve import DummyValve
from .dummy_sensor import DummySensor
from .dummy_tempcontrol import DummyTempControl



import pytest
from flowchem.components.properties import Component
from flowchem import DeviceGraph
a, b, c, d = [Component() for _ in range(4)]


@pytest.fixture
def device_graph():
    return DeviceGraph()


def test_add_single(device_graph):
    device_graph.add_device(a)
    # Contains
    assert a in device_graph
    # Length
    assert len(device_graph) == 1
    # Get by name
    assert device_graph[a.name] == a
    # Get by type
    assert len(device_graph[Component]) == 1
    # Get by value
    assert device_graph[a] == a


def test_add_iterable(device_graph):
    device_graph.add_device([a, b, c])
    assert len(device_graph) == 3
    assert a in device_graph
    assert b in device_graph
    assert c in device_graph


def test_add_errors(device_graph):

    # Not a component
    with pytest.raises(AssertionError):
        not_a_component = 5
        device_graph.add_device(not_a_component)

    # Class instead of instance
    with pytest.raises(AssertionError):
        device_graph.add_device(Component)


def test_add_edge(device_graph):
    device_graph.add_device([a, b])
    device_graph.add_connection(a, b)
    assert len(device_graph) == 2
    assert device_graph.validate()


def test_validation(device_graph):
    device_graph.add_device([a, b, c])
    device_graph.add_connection(a, b)
    # C is not connected to anything
    assert device_graph.validate() is False
    # Now DiGraph is weakly connected
    device_graph.add_connection(b, c)
    assert device_graph.validate() is True



"""" Run with uvicorn main:app """
from pathlib import Path
from loguru import logger
from typing import Dict

import yaml
from fastapi import FastAPI
from fastapi.responses import HTMLResponse

import flowchem
from flowchem.core.graph.parser import (
    DEVICE_MODULES,
    get_device_class_mapper,
)
from core.graph.validation import validate_graph
from flowchem.core.graph.devicenode import DeviceNode
from flowchem.exceptions import InvalidConfiguration


def create_server_from_config(config: Dict = None, config_file: Path = None) -> FastAPI:
    """
    Based on the yaml device graph provided, creates device objects and connect to them + .

    config: Path to the yaml file with the device config or dict.
    """

    assert (
        config is not None
        and config_file is None
        or config is None
        and config_file is not None
    )

    if config_file is not None:
        with config_file.open() as stream:
            config = yaml.safe_load(stream)

    assert isinstance(config, dict)  # This is here just to make mypy happy.

    # Validate config
    validate_graph(config)

    # FastAPI server
    app = FastAPI(title="flowchem", version=flowchem.__version__)

    # Device mapper
    device_mapper = get_device_class_mapper(DEVICE_MODULES)
    logger.debug(
        f"The following device classes have been found: {device_mapper.keys()}"
    )

    # Parse list of devices and generate endpoints
    for device_name, node_config in config["devices"].items():
        # Schema validation ensures only 1 hit here
        try:
            device_class = [
                name for name in device_mapper.keys() if name in node_config
            ].pop()
        except IndexError as e:
            raise InvalidConfiguration(
                f"No class available for device '{device_name}'"
            ) from e

        # Object type
        obj_type = device_mapper[device_class]
        device_config = node_config[device_class]

        node = DeviceNode(device_config, obj_type)
        logger.debug(f"Created device <{device_name}> with config: {device_config}")

        # Add to App
        app.include_router(
            node.router, prefix=node.router.prefix, tags=node.router.tags
        )
        logger.debug(f"Router for <{device_name}> added to app!")

    return app


if __name__ == "__main__":
    myapp = create_server_from_config(config_file=Path("../graph/sample_config.yml"))

    @myapp.get("/", response_class=HTMLResponse, include_in_schema=False)
    def root():
        """Server root"""
        return "<h1>Flowchem Device Server!</h1>" "<a href='./docs/'>API Reference</a>"

    import uvicorn

    uvicorn.run(myapp, host="127.0.0.1")



from __future__ import annotations

from loguru import logger
from typing import Optional, Union, List, Iterable, Any

import networkx as nx

from flowchem.exceptions import InvalidConfiguration
from flowchem.components.properties import Component, MultiportComponentMixin
from flowchem.components.stdlib import Tube
from flowchem.assemblies import Assembly
from flowchem.units import flowchem_ureg


class DeviceGraph:
    """
    Represents the device graph.

    This borrows logic from mw.Apparatus and ChempilerGraph
    """

    _id_counter = 0

    def __init__(self, name: Optional[str] = None):
        # if given a name, then name the apparatus, else default to a sequential name
        if name is not None:
            self.name = name
        else:
            self.name = "DeviceGraph" + str(DeviceGraph._id_counter)
            DeviceGraph._id_counter += 1

        # NetworkX Multi directed graph object
        self.graph: nx.MultiDiGraph = nx.MultiDiGraph()

    def to_yml(self):
        """Convert the graph to a YAML string"""
        import yaml

        exported_graph = dict()
        exported_graph["version"] = 1.0
        exported_graph["devices"] = {device.name: device.to_yml() for device in self.graph.nodes}
        exported_graph["physical_connections"] = {device.name: device.to_yml() for device in self.graph.edges}
        # FIXME
        return yaml.dump(exported_graph)

    def add_device(self, device: Any):
        """Add a device or list of devices to the graph"""

        if isinstance(device, Iterable):
            for component in device:
                self._add_device(component)
        else:
            self._add_device(device)

    def _add_device(self, device: Union[Component, Assembly]):
        """Adds a single device to the graph"""
        assert isinstance(device, (Component, Assembly)), "Device must be a Component or a component assembly!"
        self.graph.add_node(device)
        logger.debug(f"Added device <{device.name}> to the device graph {self.name}")

    def add_connection(
        self,
        origin: Union[str, Component],
        destination: Union[str, Component],
        origin_port: Optional[Union[str, int]] = None,
        destination_port: Optional[Union[str, int]] = None,
    ):
        """
        Add a connection to the graph, given either names or objects to be linked.

        Note: if strings are passed for origin/destination, the corresponding node MUST already be part of the graph!
        Note: if the origin or destination are Component instances that are not yet part of the graph,
              they will be added to the graph.
        """

        # If device names are passed, get the device objects
        try:
            if isinstance(origin, str):
                origin = self[origin]
                assert isinstance(origin, Component), "Origin must be a Component!"
            if isinstance(destination, str):
                destination = self[destination]
                assert isinstance(destination, Component), "Destination must be a Component!"
        except KeyError as ke:
            logger.exception(
                "A connection was attempted by node name with nodes that are not part of the graph!"
            )
            raise InvalidConfiguration("Invalid nodes for connection!") from ke

        # If ports are specified, ensure the values are valid with the respective component
        if origin_port is not None:
            assert isinstance(origin, MultiportComponentMixin), "Only MappedComponents have ports!"
            assert origin_port in origin.port, "The port specified was not found!"

        if destination_port is not None:
            assert isinstance(destination, MultiportComponentMixin), "Only MappedComponents have ports!"
            assert destination_port in destination.port, f"The port {destination_port} was not found in {destination}" \
                                                         f"[ports available are: {destination.port}]!"

        # Add the connection
        self.graph.add_edge(origin, destination, from_port=origin_port, to_port=destination_port)

    def __repr__(self):
        return f"<DeviceGraph {self.name}>"

    def __str__(self):
        return f"DeviceGraph {self.name} with {len(self)} devices."

    def __len__(self):
        return self.graph.number_of_nodes()

    def __contains__(self, item):
        try:
            self[item]
        except KeyError:
            return False
        return True

    def __getitem__(self, item):
        """
        Utility method

        DeviceGraph['name'] gives the device with that name
        DeviceGraph[class] returns a list of devices of that type
        DeviceGraph[device_instance] returns true if the object is part of the graph
        """
        # If a type is passed return devices with that type
        if isinstance(item, type):
            return [device for device in self.graph.nodes if isinstance(device, item)]

        # If a string is passed return the device with that name
        elif isinstance(item, str):
            for node in self.graph.nodes:
                if node.name == item:
                    return node
            raise KeyError(f"No component named '{item}' in {repr(self)}.")

        # a shorthand way to check if a component is in the apparatus
        elif item in self.graph.nodes:
            return item
        else:
            raise KeyError(f"{repr(item)} is not in {repr(self)}.")

    def component_from_origin_and_port(self, origin: Component, port: Union[str, int]) -> Component:
        """ Returns the component that is connected to the origin component in the port provided. """

        assert origin in self, "The origin component is not part of the graph!"
        assert isinstance(origin, MultiportComponentMixin), "Only MappedComponents have ports!"

        for _, to_component, data in self.graph.out_edges(origin, data=True):
            if data["from_port"] == port:
                return to_component
        raise KeyError(f"No component connected to port {port}")

    def visualize(self):
        """Visualize the graph"""
        import matplotlib.pyplot as plt

        nx.draw(self.graph, with_labels=True)
        plt.show()

    def explode_all(self):
        """Explode all devices in the graph"""
        # Copy list of nodes to prevent change during iteration
        original_nodes = list(self.graph.nodes)
        for device in original_nodes:
            # if a device has an 'explode' method it means it is a device group.
            # Note: this naive approach would fail on nested groups
            if hasattr(device, "explode"):
                logger.debug(f"Exploding {device.name}")
                device.explode(self)

    def validate(self) -> bool:
        """Validates the graph. This is called by Protocol when the DeviceGraph is used."""

        # Make sure that all the components are connected
        return nx.is_weakly_connected(self.graph)

    def summarize(self):
        """
        Prints a summary table of the DeviceGraph.
        Rich takes care of the formatting both in console and jupyter cells.
        """
        from rich.table import Table

        # Components table
        components_table = Table(title=f"Components")

        # Columns: Name, Type
        components_table.add_column("Name")
        components_table.add_column("Type")

        # Fill rows with all devices while skipping tubes (saving them for the second table)
        tubes: List[Tube] = []
        for component in sorted(self.graph.nodes, key=lambda x: x.__class__.__name__):
            if component.__class__.__name__ != "Tube":
                components_table.add_row(component.name, component.__class__.__name__)
            else:
                tubes.append(component)

        # Tubing table
        tubing_table = Table(
            "From", "To", "Length", "I.D.", "O.D.", "Volume", "Material", title="Tubing"
        )

        # store and calculate the computed totals for tubing
        total_length = 0 * flowchem_ureg.mm
        total_volume = 0 * flowchem_ureg.ml

        last_tube = tubes[-1]
        for tube in tubes:
            total_length += tube.length
            total_volume += tube.volume
            from_component = next(self.graph.predecessors(tube))
            to_component = next(self.graph.successors(tube))

            # Draw a line after the last tube
            end_section = True if tube is last_tube else False

            tubing_table.add_row(
                from_component.name,
                to_component.name,
                f"{tube.length:~H}",
                f"{tube.ID:~H}",
                f"{tube.OD:~H}",
                f"{tube.volume.to('ml'):.4f~H}",
                tube.material,
                end_section=end_section,
            )

        tubing_table.add_row(
            "Total",
            "n/a",
            f"{total_length:~H}",
            "n/a",
            "n/a",
            f"{total_volume.to('ml'):.4f~H}",
            "n/a",
        )

        # Print tables
        from rich.console import Console

        console = Console()
        console.print(components_table)
        console.print(tubing_table)


if __name__ == "__main__":
    from flowchem.core.graph.parser import parse_graph_file
    graph = parse_graph_file("sample_config.yml")
    graph.summarize()

    graph.explode_all()
    graph.summarize()

    # from flowchem import Protocol
    # a = graph.to_apparatus()
    # print(a)
    # p = Protocol(a)
    #
    # from datetime import timedelta
    # t0 = timedelta(seconds=0)
    #
    # # p.add(graph["quencher"], start=t0, duration=timedelta(seconds=10), rate="0.1 ml/min")
    # p.add(
    #     graph["activator"], start=t0, duration=timedelta(seconds=10), rate="0.1 ml/min"
    # )
    # print(graph["chiller"])
    # print(type(graph["chiller"]))
    # p.add(graph["chiller"], start=t0, duration=timedelta(seconds=10), temp="45 degC")
    #
    # E = p.execute(dry_run=False)
    # # E.visualize()



""" For each device node described in the configuration [graph] instantiated it and create endpoints """
import inspect
import warnings

from loguru import logger
from fastapi import APIRouter

from flowchem.components.properties import ActiveComponent
from flowchem import Spinsolve
from flowchem.core.server.routers import spinsolve_get_router


class DeviceNode:
    """Represent a node in the device graph, holds the HW object and its metadata/config."""

    # Router generators for device class that do not implement self.get_router()
    # All callable take the device obj and return an APIRouter
    router_generator = {Spinsolve: spinsolve_get_router}

    def __init__(self, device_config, obj_type):
        self._router = None

        # No configuration for t-mixer et al.
        if device_config is None:
            device_config = {}

        # Ensure the name is set
        if "name" not in device_config:
            warnings.warn("Device name not set, using class name")
            device_config["name"] = obj_type.__name__

        # DEVICE INSTANTIATION
        try:
            # Special class method for initialization required for some devices
            if hasattr(obj_type, "from_config"):
                self.device = obj_type.from_config(**device_config)
            else:
                self.device = obj_type(**device_config)
            logger.debug(f"Created {self.device.name} instance: {self.device}")
        except TypeError as e:
            raise ConnectionError(
                f"Wrong configuration provided for device: {device_config.get('name')} of type {obj_type}!\n"
                f"Configuration: {device_config}\n"
                f"Accepted parameters: {inspect.getfullargspec(obj_type).args}"
            ) from e

    @property
    def router(self):
        """Returns an APIRouter associated with the device"""
        if self._router:
            return self._router

        if hasattr(self.device, "get_router"):
            router = self.device.get_router()
        else:
            try:
                router = DeviceNode.router_generator[type(self.device)](self.device)
            except KeyError:
                # Only warn no router for active components
                if isinstance(self.device, ActiveComponent):
                    logger.warning(
                        f"No router available for device '{self.device.name}'"
                        f"[Class: {type(self.device).__name__}]"
                    )
                router = APIRouter()

        """ Router name is lowercase with no whitespace """
        router_name = self.device.name.replace(" ", "").lower()

        router.prefix = f"/{router_name}"
        router.tags = [router_name]
        self._router = router
        return self._router




from flowchem.components.dummy import DummyPump, DummySensor, BrokenDummySensor
from flowchem.components.stdlib import Vessel, Tube
from flowchem import Protocol, DeviceGraph

# create components
from flowchem.units import flowchem_ureg

#
# def test_execute():
#     a = Vessel(name="a", description="nothing")
#     b = Vessel(name="b", description="nothing")
#     c = Vessel(name="c", description="nothing")
#
#     pump = DummyPump(name="Dummy pump")
#
#     test = DummySensor(name="test")
#     test2 = DummySensor(name="test2")
#     test3 = DummySensor(name="test3")
#     test4 = BrokenDummySensor(name="test4")
#
#     tube = Tube("1 foot", "1/16 in", "2/16 in", "PVC")
#
#     # create apparatus
#     D = DeviceGraph()
#     # A = Apparatus()
#     A.add([a, b, c], pump, tube)
#     A.add(pump, [test, test2, test3, test4], tube)
#
#     P = Protocol(A, name="testing execution")
#     P.add(pump, rate="5 mL/min", start="0 seconds", stop="1 secs")
#     P.add([test, test2, test3, test4], rate="5 Hz", start="0 secs", stop="1 secs")
#
#     # test both execution modes
#     for dry_run in [True, False]:
#         E = P.execute(confirm=True, dry_run=dry_run, log_file=None, data_file=None)
#
#         assert len(E.data["test"]) >= 5
#         if dry_run:
#             assert E.data["test"][0].data == "simulated read"
#         assert pump.rate == flowchem_ureg.parse_expression(pump._base_state["rate"])
#
#     # test fast forward
#     E = P.execute(confirm=True, dry_run=5, log_file=None, data_file=None)
#     assert len(E.data["test"]) >= 1



import pytest

from flowchem.components.stdlib import Tube


def test_diameter_validation():
    # inner diameter is greater than outer diameter
    with pytest.raises(ValueError):
        Tube(length="5 cm", ID="2 cm", OD="1 cm", material="boyfriend material")


def test_length_units():
    with pytest.raises(ValueError):
        Tube(length="5 L", ID="1 cm", OD="3 cm", material="boyfriend material")


def test_repr():
    assert (
        repr(Tube(length="5 cm", ID="1 cm", OD="2 cm", material="boyfriend material"))
        == "Tube of length 5 centimeter, ID 1 centimeter, OD 2 centimeter"
    )



""" Test HuberChiller object. Does not require physical connection to the device. """
import asyncio

import aioserial
import pytest

from flowchem.components.devices.Huber.huberchiller import HuberChiller, PBCommand
from flowchem.exceptions import InvalidConfiguration


# TEST PBCommand parsers first
def test_pbcommand_parse_temp():
    assert PBCommand("{S00F2DF").parse_temperature() == "-33.61 degree_Celsius"
    assert PBCommand("{S0004DA").parse_temperature() == "12.42 degree_Celsius"


def test_pbcommand_parse_int():
    assert PBCommand("{S000000").parse_integer() == 0
    assert PBCommand("{S00ffff").parse_integer() == 65535
    assert PBCommand("{S001234").parse_integer() == 4660


def test_pbcommand_parse_bits():
    assert PBCommand("{S001234").parse_bits() == [False, False, False, True, False, False, True, False, False, False,
                                                  True, True, False, True, False, False]


def test_pbcommand_parse_bool():
    assert PBCommand("{S000001").parse_boolean() is True
    assert PBCommand("{S000000").parse_boolean() is False


def test_invalid_serial_port():
    with pytest.raises(InvalidConfiguration) as execinfo:
        HuberChiller.from_config(port="COM99")
    assert str(execinfo.value) == 'Cannot connect to the HuberChiller on the port <COM99>'


class FakeSerial(aioserial.AioSerial):
    """ Mock AioSerial. """

    # noinspection PyMissingConstructor
    def __init__(self):
        self.fixed_reply = None
        self.last_command = b""
        self.map_reply = {
            b"{M0A****\r\n": b"{S0AFFFF\r\n",  # Fake status reply
            b"{M3C****\r\n": b"{S3CFFFF\r\n",  # Fake status2 reply
            b"{M00****\r\n": b"{S0004DA\r\n",  # Fake setpoint reply
            b"{M03****\r\n": b"{S030a00\r\n",  # Fake pressure reply
            b"{M04****\r\n": b"{S04000a\r\n",  # Fake current power reply (10 W)
            b"{M26****\r\n": b"{S26000a\r\n",  # Fake current pump speed (10 rpm)
            b"{M30****\r\n": b"{S30EC78\r\n",  # Fake min temp -50.00 C
            b"{M31****\r\n": b"{S303A98\r\n",  # Fake max temp +150.00 C
            b"{M00EC78\r\n": b"{S00EC78\r\n",  # set temp to -50
            b"{M003A98\r\n": b"{S003A98\r\n",  # set temp to 150
            b"{M0007D0\r\n": b"{S0007D0\r\n",  # Reply to set temp 20 C
            b"{M00F830\r\n": b"{S00F830\r\n",  # Reply to set temp -20 C
        }

    async def write_async(self, text: bytes):
        """ Override AioSerial method """
        self.last_command = text

    async def readline_async(self, size: int = -1) -> bytes:
        """ Override AioSerial method """
        if self.last_command == b"{MFFFFFF\r\n":
            await asyncio.sleep(999)
        if self.fixed_reply:
            return self.fixed_reply
        return self.map_reply[self.last_command]

    def __repr__(self):
        return "FakeSerial"


@pytest.fixture(scope="session")
def chiller():
    """ Chiller instance connected to FakeSerial """
    return HuberChiller(FakeSerial())


@pytest.mark.asyncio
async def test_no_reply(chiller):
    with pytest.warns(UserWarning):
        reply = await chiller.send_command_and_read_reply("{MFFFFFF")
    assert reply == '{SFFFFFF'


@pytest.mark.asyncio
async def test_status(chiller):
    chiller._serial.fixed_reply = None
    stat = await chiller.status()
    stat_content = [x for x in stat.values()]
    assert all(stat_content)

    # Set reply in FakeSerial
    chiller._serial.fixed_reply = b"{S0A0000"
    stat = await chiller.status()
    stat_content = [x for x in stat.values()]
    assert not any(stat_content)


@pytest.mark.asyncio
async def test_status2(chiller):
    chiller._serial.fixed_reply = None
    stat = await chiller.status2()
    stat_content = [x for x in stat.values()]
    assert all(stat_content)

    # Set reply in FakeSerial
    chiller._serial.fixed_reply = b"{S0A0000"
    stat = await chiller.status2()
    stat_content = [x for x in stat.values()]
    assert not any(stat_content)


@pytest.mark.asyncio
async def test_get_temperature_setpoint(chiller):
    chiller._serial.fixed_reply = None
    temp = await chiller.get_temperature_setpoint()
    assert temp == "12.42 degree_Celsius"

    chiller._serial.fixed_reply = b"{S00F2DF"
    temp = await chiller.get_temperature_setpoint()
    assert temp == "-33.61 degree_Celsius"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_temperature_setpoint(chiller):
    chiller._serial.fixed_reply = None
    await chiller.set_temperature_setpoint("20 °C")
    print(chiller._serial.last_command)
    assert chiller._serial.last_command == b"{M0007D0\r\n"

    await chiller.set_temperature_setpoint("-20 °C")
    assert chiller._serial.last_command == b"{M00F830\r\n"

    with pytest.warns(Warning):
        await chiller.set_temperature_setpoint("-400 °C")
        assert chiller._serial.last_command == b"{M00EC78\r\n"

    with pytest.warns(Warning):
        await chiller.set_temperature_setpoint("4000 °C")
        assert chiller._serial.last_command == b"{M003A98\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_internal_temperature(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.internal_temperature()
    assert chiller._serial.last_command == b"{M01****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_return_temperature(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.return_temperature()
    assert chiller._serial.last_command == b"{M02****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_process_temperature(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.process_temperature()
    assert chiller._serial.last_command == b"{M3A****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_pump_pressure(chiller):
    chiller._serial.fixed_reply = None
    pressure = await chiller.pump_pressure()
    assert chiller._serial.last_command == b"{M03****\r\n"
    assert pressure == "2560 millibar"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_current_power(chiller):
    chiller._serial.fixed_reply = None
    power = await chiller.current_power()
    assert chiller._serial.last_command == b"{M04****\r\n"
    assert power == "10 watt"


@pytest.mark.asyncio
async def test_get_temperature_control(chiller):
    chiller._serial.fixed_reply = b"{S140000"
    t_ctl = await chiller.is_temperature_control_active()
    assert t_ctl is False
    chiller._serial.fixed_reply = b"{S140001"
    t_ctl = await chiller.is_temperature_control_active()
    assert t_ctl is True


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_temperature_control(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.start_temperature_control()
    assert chiller._serial.last_command == b"{M140001\r\n"
    await chiller.stop_temperature_control()
    assert chiller._serial.last_command == b"{M140000\r\n"


@pytest.mark.asyncio
async def test_get_circulation(chiller):
    chiller._serial.fixed_reply = b"{S160000"
    circulation = await chiller.is_circulation_active()
    assert circulation is False
    chiller._serial.fixed_reply = b"{S160001"
    circulation = await chiller.is_circulation_active()
    assert circulation is True


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_circulation(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.start_circulation()
    assert chiller._serial.last_command == b"{M160001\r\n"
    await chiller.stop_circulation()
    assert chiller._serial.last_command == b"{M160000\r\n"


@pytest.mark.asyncio
async def test_min_setpoint(chiller):
    chiller._serial.fixed_reply = None
    min_t = await chiller.min_setpoint()
    assert min_t == "-50.0 degree_Celsius"


@pytest.mark.asyncio
async def test_max_setpoint(chiller):
    chiller._serial.fixed_reply = None
    max_t = await chiller.max_setpoint()
    assert max_t == "150.0 degree_Celsius"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_pump_speed(chiller):
    chiller._serial.fixed_reply = None
    speed = await chiller.pump_speed()
    assert chiller._serial.last_command == b"{M26****\r\n"
    assert speed == "10 revolutions_per_minute"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_pump_speed_setpoint(chiller):
    chiller._serial.fixed_reply = b"{S480000"
    speed = await chiller.pump_speed_setpoint()
    assert chiller._serial.last_command == b"{M48****\r\n"
    assert speed == "0 revolutions_per_minute"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_pump_speed(chiller):
    chiller._serial.fixed_reply = b"{S480000"
    await chiller.set_pump_speed("10 rpm")
    assert chiller._serial.last_command == b"{M48000A\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_cooling_water_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.cooling_water_temp()
    assert chiller._serial.last_command == b"{M2C****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_cooling_water_pressure(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.cooling_water_pressure()
    assert chiller._serial.last_command == b"{M2D****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_cooling_water_temp_out(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.cooling_water_temp_outflow()
    assert chiller._serial.last_command == b"{M4C****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_alarm_max_internal_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.alarm_max_internal_temp()
    assert chiller._serial.last_command == b"{M51****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_alarm_min_internal_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.alarm_min_internal_temp()
    assert chiller._serial.last_command == b"{M52****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_alarm_max_process_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.alarm_max_process_temp()
    assert chiller._serial.last_command == b"{M53****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_alarm_min_process_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.alarm_min_process_temp()
    assert chiller._serial.last_command == b"{M54****\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_alarm_max_internal_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.set_alarm_max_internal_temp("10 °C")
    assert chiller._serial.last_command == b"{M5103E8\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_alarm_min_internal_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.set_alarm_min_internal_temp("10 °C")
    assert chiller._serial.last_command == b"{M5203E8\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_alarm_max_process_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.set_alarm_max_process_temp("10 °C")
    assert chiller._serial.last_command == b"{M5303E8\r\n"


# noinspection PyUnresolvedReferences
@pytest.mark.asyncio
async def test_set_alarm_min_process_temp(chiller):
    chiller._serial.fixed_reply = b"{S000000"
    await chiller.set_alarm_min_process_temp("10 °C")
    assert chiller._serial.last_command == b"{M5403E8\r\n"



"""
Knauer pump
Run with python -m pytest ./tests -m KPump and updates pump address below
"""
import asyncio
import math
import sys

import pint
import pytest

from flowchem import AzuraCompactPump
from flowchem.components.devices.Knauer.AzuraCompactPump import AzuraPumpHeads

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


# noinspection PyUnusedLocal
@pytest.fixture(scope="session")
def event_loop(request):
    """

    Args:
        request:
    """
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
async def pump():
    """ Change to match your hardware ;) """
    pump = AzuraCompactPump(ip_address="192.168.1.126")
    await pump.initialize()
    return pump


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_pumphead(pump: AzuraCompactPump):
    assert await pump.get_headtype() in AzuraPumpHeads
    await pump.set_headtype(AzuraPumpHeads.FLOWRATE_TEN_ML)
    assert await pump.get_headtype() == AzuraPumpHeads.FLOWRATE_TEN_ML
    await pump.set_headtype(AzuraPumpHeads.FLOWRATE_FIFTY_ML)
    assert await pump.get_headtype() == AzuraPumpHeads.FLOWRATE_FIFTY_ML


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_headtype(pump: AzuraCompactPump):
    assert await pump.get_headtype() in AzuraPumpHeads


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_flow_rate(pump: AzuraCompactPump):
    await pump.set_flow("1.25 ml/min")
    await pump.start_flow()
    # FIXME
    assert pint.Quantity(await pump.get_flow()).magnitude == 1.25
    await pump.set_flow(f"{math.pi} ml/min")
    assert math.isclose(pint.Quantity(await pump.get_flow()).magnitude, math.pi, abs_tol=1e-3)
    await pump.stop_flow()


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_analog_control(pump: AzuraCompactPump):
    await pump.enable_analog_control(True)
    assert await pump.is_analog_control_enabled() is True
    await pump.enable_analog_control(False)
    assert await pump.is_analog_control_enabled() is False


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_is_running(pump: AzuraCompactPump):
    await pump.set_flow("1 ml/min")
    await pump.start_flow()
    assert pump.is_running() is True
    await pump.stop_flow()


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_motor_current(pump: AzuraCompactPump):
    await pump.stop_flow()
    assert await pump.read_motor_current() == 0


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_correction_factor(pump: AzuraCompactPump):
    init_val = await pump.get_correction_factor()
    await pump.set_correction_factor(0)
    assert await pump.get_correction_factor() == 0
    await pump.set_correction_factor(init_val)


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_adjusting_factor(pump: AzuraCompactPump):
    init_val = await pump.get_adjusting_factor()
    await pump.set_adjusting_factor(1000)
    assert await pump.get_adjusting_factor() == 1000
    await pump.set_adjusting_factor(init_val)


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_autostart(pump: AzuraCompactPump):
    await pump.enable_autostart()
    assert await pump.is_autostart_enabled() is True
    await pump.enable_autostart(False)


@pytest.mark.KPump
@pytest.mark.asyncio
async def test_start_in(pump: AzuraCompactPump):
    await pump.require_start_in()
    assert await pump.is_start_in_required() is True
    await pump.require_start_in(False)



import yaml
from pathlib import Path
from typing import Union, Dict, Iterable
from loguru import logger
from types import ModuleType
import inspect
import itertools

from flowchem.components.stdlib import Tube
from flowchem.core.graph.devicenode import DeviceNode
from flowchem.core.graph.validation import validate_graph
from flowchem.exceptions import InvalidConfiguration
from flowchem.core.graph.devicegraph import DeviceGraph
import flowchem.assemblies


# Packages containing the device class definitions.
# Devices' classes must be in the module top level to be found.
DEVICE_MODULES = [
    flowchem.components.devices,
    flowchem.components.stdlib,
    flowchem.components.dummy,
    flowchem.assemblies,
]


def get_device_class_mapper(modules: Iterable[ModuleType]) -> Dict[str, type]:
    """
    Given an iterable of modules containing the device classes, return a
    dictionary Dict[device_class_name, DeviceClass]

    Args:
        modules (Iterable[ModuleType]): The modules to inspect for devices.
            Only class in the top level of each module will be extracted.
    Returns:
        device_dict (Dict[str, type]): Dict of device class names and their
            respective classes, i.e. {device_class_name: DeviceClass}.
    """
    # Get (name, obj) tuple for the top level of each module.
    objects_in_modules = [
        inspect.getmembers(module, inspect.isclass) for module in modules
    ]

    # Return them as dict (itertools to flatten the nested, per module, lists)
    return {k: v for (k, v) in itertools.chain.from_iterable(objects_in_modules)}


def parse_device_section(devices: Dict, graph: DeviceGraph):
    """Parse the devices' section of the graph config"""

    # Device mapper needed for device instantiation
    device_mapper = get_device_class_mapper(DEVICE_MODULES)
    logger.debug(f"Device classes found: {device_mapper.keys()}")

    # Parse devices
    for device_node in devices:
        device_class, device_config = next(iter(device_node.items()))
        try:
            obj_type = device_mapper[device_class]
        except KeyError as e:
            logger.exception(f"Device of type {device_class} unknown! [Known devices: {device_mapper.keys()}]")
            raise InvalidConfiguration(f"Device of type {device_class} unknown! \n"
                                       f"[Known devices: {list(device_mapper.keys())}]") from e

        # Create device object and add it to the graph
        device = DeviceNode(device_config, obj_type).device
        graph.add_device(device)


def parse_connection_section(connections: Dict, graph: DeviceGraph):
    """Parse connections from the graph config"""
    for edge in connections:
        edge_class, edge_config = next(iter(edge.items()))

        if "Tube" in edge_class:
            _parse_tube_connection(edge_config, graph)
        elif "Interface" in edge_class:
            _parse_interface_connection(edge_config, graph)
        else:
            raise InvalidConfiguration(f"Invalid connection type in {edge}")


def _parse_tube_connection(tube_config, graph: DeviceGraph):
    """
    The Tube object is a convenience object for connecting devices without explicitly creating the
    in-between tube node.
    """
    tube = Tube(
        length=tube_config["length"],
        ID=tube_config["inner-diameter"],
        OD=tube_config["outer-diameter"],
        material=tube_config["material"],
    )
    graph.add_device(tube)

    # Create logic connections for newly created tube
    inlet = {
        "from": dict(
            device=tube_config["from"]["device"],
            port=tube_config["from"].get("port", None)),
        "to": dict(device=tube.name),
    }
    _parse_interface_connection(inlet, graph)

    outlet = {
        "from": dict(device=tube.name),
        "to": dict(
            device=tube_config["to"]["device"],
            port=tube_config["to"].get("port", None)),
    }
    _parse_interface_connection(outlet, graph)


def _parse_interface_connection(iface_config, graph: DeviceGraph):
    """Parse a dict containing the Tube connection and returns the Connection"""
    graph.add_connection(
        origin=iface_config["from"]["device"],
        destination=iface_config["to"]["device"],
        origin_port=iface_config["from"].get("port", None),
        destination_port=iface_config["to"].get("port", None),
    )


def parse_graph_config(graph_config: Dict, name: str = None) -> DeviceGraph:
    """Parse a graph config and returns a DeviceGraph object."""

    # Validate graph
    validate_graph(graph_config)

    # Create DeviceGraph object
    device_graph = DeviceGraph(name)

    # Parse devices
    parse_device_section(graph_config["devices"], device_graph)

    # Parse connections
    parse_connection_section(graph_config["connections"], device_graph)

    logger.info(f"Parsed graph {name}")
    return device_graph


def parse_graph_file(file: Union[str, Path]):
    """Parse a graph config file and returns a DeviceGraph object."""
    file_path = Path(file)
    name = file_path.stem

    with file_path.open() as stream:
        config = yaml.safe_load(stream)

    return parse_graph_config(config, name)



""" Validate graph files """
from __future__ import annotations
from typing import Dict

import os
import json
import jsonschema

# Validation schema for graph file
SCHEMA = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../graph/flowchem-graph-spec.schema"
)


def load_graph_schema():
    """loads the schema defining valid config file."""
    with open(SCHEMA, "r") as fp:
        schema = json.load(fp)
        jsonschema.Draft7Validator.check_schema(schema)
        return schema


def validate_graph(graph: Dict):
    schema = load_graph_schema()
    jsonschema.validate(graph, schema=schema)
    assert graph["version"] == "1.1"



""" Unit-conversion related functions """
import pint

flowchem_ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)
flowchem_ureg.define("step = []")
flowchem_ureg.define("stroke = 48000 * step")



from typing import Tuple, Sequence, TYPE_CHECKING

from flowchem.components.properties import MultiportComponentMixin
from flowchem.components.properties import Component

if TYPE_CHECKING:
    from flowchem.core.graph import DeviceGraph


class Assembly(MultiportComponentMixin, Component):
    """ A class representing a collection of components. """
    nodes: Sequence[Component]
    edges: Sequence[Tuple[Component, Component]]

    def _subcomponent_by_name(self, name: str) -> Component:
        """ Returns a component in nodes by its name. """
        for node in self.nodes:
            if node.name == name:
                return node
        raise ValueError(f"No component named {name} in {self}")

    def explode(self, graph: "DeviceGraph"):
        """
        Explode the assembly into its components in the provided graph.
        The graph must already include the assembly as a node with all the connections defined.
        """

        assert self in graph.graph.nodes, "Assembly must be in the graph to explode it."

        # Convert edges to self into edges to self's components.
        for from_component, to_component, attributes in graph.graph.in_edges(self, data=True):
            # If port attribute is unspecified, the connection is assumed to all the assembly's subcomponents.
            # This should only happen for logical connections (e.g. temp control).
            if attributes["to_port"] is None:
                for subcomponent in self.nodes:
                    graph.graph.add_edge(from_component, subcomponent)
                continue

            # New destination is the component with name matching the edge port on the assembly
            new_to_component = self._subcomponent_by_name(attributes["to_port"])

            # Update edge - just add a new one, the old one will be implicitly removed with graph.remove_node(self)
            graph.add_connection(origin=from_component, destination=new_to_component,
                                 origin_port=attributes.get("from_port", None))

        for from_component, to_component, attributes in graph.graph.out_edges(self, data=True):
            assert from_component is self, "Getting the edges pointing from the assembly."

            # New origin is the component with name matching the edge port on the assembly
            new_from_component = self._subcomponent_by_name(attributes["from_port"])

            # Update edge - just add a new one, the old one will be implicitly removed with graph.remove_node(self)
            graph.add_connection(origin=new_from_component, destination=to_component,
                                 destination_port=attributes.get("to_port", None))

        # Updates component names. Ensures unique names in the graph. (Note: do not update those earlier: see above!)
        for component in self.nodes:
            component.name = f"{self.name}_{component.name}"

        # Remove assembly from graph (this also removes all edges)
        graph.graph.remove_node(self)

        # Add nodes to graph
        graph.add_device(self.nodes)
        # Add edges to graph
        for edge in self.edges:
            graph.add_connection(edge[0], edge[1])

    def _validate(self, dry_run):
        """Components are valid for dry runs, but not for real runs."""
        raise NotImplementedError("Assembly object should be expanded into their components before run.")



""" Test Spinsolve, needs actual connection with the device """
import pytest
import asyncio
import time
from pathlib import Path
from flowchem.components.devices.Magritek.spinsolve import Spinsolve


host = "BSMC-7WP43Y1"


@pytest.fixture(scope="session")
def nmr():
    """ Spinsolve instance on host:port. Change to match your hardware ;) """
    return Spinsolve.from_config(host=host)


@pytest.mark.Spinsolve
def test_connection(nmr: Spinsolve):
    assert isinstance(nmr, Spinsolve)


@pytest.mark.Spinsolve
def test_solvent_property(nmr: Spinsolve):
    test_string = "test-solvent"
    nmr.solvent = test_string
    time.sleep(0.2)  # Ensures property is set.
    assert nmr.solvent == test_string


@pytest.mark.Spinsolve
def test_sample_property(nmr: Spinsolve):
    test_string = "sample_name"
    nmr.sample = test_string
    time.sleep(0.2)  # Ensures property is set.
    assert nmr.sample == test_string


@pytest.mark.Spinsolve
def test_user_data(nmr: Spinsolve):
    # Assignment is actually addition
    nmr.user_data = dict(key1="value1")
    time.sleep(0.2)  # Ensures property is set.
    assert "key1" in nmr.user_data
    nmr.user_data = dict(key2="value2")
    time.sleep(0.2)  # Ensures property is set.
    assert "key1" in nmr.user_data
    assert "key2" in nmr.user_data
    # Removal obtained by providing empty strings
    nmr.user_data = dict(key1="", key2="")
    time.sleep(0.2)  # Ensures property is set.
    assert "key1" not in nmr.user_data
    assert "key2" not in nmr.user_data


@pytest.mark.Spinsolve
def test_hw_request(nmr: Spinsolve):
    hw_tree = nmr.hw_request()
    assert hw_tree.find(".//SpinsolveSoftware") is not None


@pytest.mark.Spinsolve
def test_request_available_protocols(nmr: Spinsolve):
    protocols = nmr.request_available_protocols()
    assert isinstance(protocols, dict)
    assert "1D PROTON" in protocols


@pytest.mark.Spinsolve
def test_is_protocol_available(nmr: Spinsolve):
    assert nmr.is_protocol_available("1D PROTON")


@pytest.mark.Spinsolve
def test_request_validation(nmr: Spinsolve):
    # VALID
    valid_protocol = dict(Number="8", AcquisitionTime="3.2", RepetitionTime="2", PulseAngle="45")
    check_protocol = nmr._validate_protocol_request("1D EXTENDED+", valid_protocol)
    assert check_protocol == valid_protocol

    # INVALID NAME
    check_protocol = nmr._validate_protocol_request("NON EXISTING PROTOCOL", valid_protocol)
    assert not check_protocol

    # PARTLY VALID OPTIONS
    partly_valid = dict(Number="7", AcquisitionTime="3.2", RepetitionTime="2", PulseAngle="145")
    with pytest.warns(UserWarning, match="Invalid value"):
        check_protocol = nmr._validate_protocol_request("1D EXTENDED+", partly_valid)
    assert check_protocol == dict(AcquisitionTime="3.2", RepetitionTime="2")

    # INVALID OPTIONS 1
    partly_valid = dict(Number="7", AcquisitionTime="43.2", RepetitionTime="2123092183", PulseAngle="145")
    with pytest.warns(UserWarning, match="Invalid value"):
        check_protocol = nmr._validate_protocol_request("1D EXTENDED+", partly_valid)
    assert not check_protocol

    # INVALID OPTIONS 21
    partly_valid = dict(Number="8", AcquisitionTime="3.2", RepetitionTime="2", PulseAngle="45", blabla="no")
    with pytest.warns(UserWarning, match="Invalid option"):
        check_protocol = nmr._validate_protocol_request("1D EXTENDED+", partly_valid)
    assert "balbla" not in check_protocol


@pytest.mark.Spinsolve
def test_data_folder(nmr: Spinsolve):
    nmr.data_folder = "C:/data/myfolder"
    time.sleep(0.1)
    assert nmr.data_folder == "C:/data/myfolder"
    nmr.data_folder = "C:/data/myfolder"
    time.sleep(0.1)
    assert nmr.data_folder == "C:/data/myfolder"
    nmr.data_folder = "C:/data/myfolder"
    time.sleep(0.1)
    assert nmr.data_folder == "C:/data/myfolder"


@pytest.mark.Spinsolve
def test_protocol(nmr: Spinsolve):
    # back to default
    nmr.data_folder = ""
    time.sleep(0.1)

    # Run fast proton
    path = asyncio.run(nmr.run_protocol("1D PROTON", dict(Scan="QuickScan")))
    assert isinstance(path, Path)


@pytest.mark.Spinsolve
def test_invalid_protocol(nmr: Spinsolve):
    # back to default
    nmr.data_folder = ""
    time.sleep(0.1)

    # Fail on plutonium
    with pytest.warns(UserWarning, match="not available"):
        path = asyncio.run(nmr.run_protocol("1D PLUTONIUM", dict(Scan="QuickScan")))
    assert path is None



import pytest

from flowchem.components.properties import Component, ActiveComponent, Sensor
from flowchem.units import flowchem_ureg


def test_validate_component():
    # not a subclass of ActiveComponent, so it won't work
    class Test(Component):
        def __init__(self):
            super().__init__()

    Test()._validate(dry_run=True)
    with pytest.raises(RuntimeError):
        Test()._validate(dry_run=False)


def test_empty_base_state():
    # base_state dictionary is not valid for the component because it's empty
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.active = False
            self._base_state = {}

        async def _update(self):
            return True

    with pytest.raises(ValueError):
        Test()._validate(dry_run=True)


def test_invalid_base_state():
    # base_state dictionary is not valid for the component
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.active = False
            self._base_state = dict(rate="10 mL")

        async def _update(self):
            return True

    with pytest.raises(ValueError):
        Test()._validate(dry_run=True)


def test_wrong_base_state_dimensionality():
    # base_state dictionary is wrong dimensionality
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.rate = flowchem_ureg.parse_expression("10 mL/min")
            self._base_state = dict(rate="10 mL")

        async def _update(self):
            return True

    with pytest.raises(ValueError):
        Test()._validate(dry_run=True)


def test_passing_class():
    class Test(ActiveComponent):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.active = False
            self.serial_port = serial_port
            self._base_state = dict(active=False)

        async def _update(self):
            pass

    # should pass both as a dry run and as a real run (since update doesn't do anything)
    Test()._validate(dry_run=True)
    Test()._validate(dry_run=False)


def test_base_state_type():
    # not right base_state value type
    class Test(ActiveComponent):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.active = False
            self.serial_port = serial_port
            self._base_state = dict(active="10 mL")

        async def _update(self):
            return True

    with pytest.raises(ValueError):
        Test()._validate(dry_run=True)

    # not right base_state value type
    class Test(ActiveComponent):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.active = False
            self.serial_port = serial_port
            self._base_state = "not a dict"

        async def _update(self):
            return True

    with pytest.raises(ValueError):
        Test()._validate(dry_run=True)


def test_validate_sensor_without_read():
    class Test(Sensor):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.serial_port = serial_port

    with pytest.raises(NotImplementedError):
        Test()._validate(dry_run=False)
    Test()._validate(dry_run=True)  # should pass during a dry run


def test_update_must_return_none():
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.active = False
            self._base_state = dict(active=False)

        async def _update(self):
            return False

    Test()._validate(dry_run=True)
    with pytest.raises(ValueError):
        Test()._validate(dry_run=False)


def test_default_update():
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.active = False
            self._base_state = dict(active=False)

    Test()._validate(dry_run=True)
    with pytest.raises(NotImplementedError):
        Test()._validate(dry_run=False)


def test_sync_update():
    class Test(ActiveComponent):
        def __init__(self):
            super().__init__(name=None)
            self.active = False
            self._base_state = dict(active=False)

        def _update(self):
            pass

    Test()._validate(dry_run=True)
    with pytest.raises(TypeError):
        Test()._validate(dry_run=False)


def test_sync_read():
    class Test(Sensor):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.serial_port = serial_port

        def _read(self):
            return 1

        async def _update(self):
            pass

    Test()._validate(dry_run=True)
    with pytest.raises(TypeError):
        Test()._validate(dry_run=False)


def test_validate_sensor_with_failing_read():
    class Test(Sensor):
        def __init__(self, serial_port=None):
            super().__init__(name=None)
            self.serial_port = serial_port

        async def _read(self):
            raise RuntimeError("This component is broken!")

        async def _update(self):
            pass

    Test()._validate(dry_run=True)
    with pytest.raises(RuntimeError):
        Test()._validate(dry_run=False)



from typing import Optional, Set, Union

from flowchem.components.properties import Component


class MultiportComponentMixin(Component):
    """
    A Mixin to be added to component with more than 1 inlet or outlet port.

    All components with multimple ports should derive from this to ensure proper validation of the port positions.

    The port names have to be provided as follows (e.g. w/ a set, ensuring unique names):
    - self.port = {'position_name_1', 'position_name_2'}
    where 'position_name_1' can be an int (e.g. multipos. valves) or a string (e.g. 'inject', 'load') for 2-pos valves.
    The port names specified will be used as edge attributes in the graph (attrs. from_position and to_position).

    A Mixin is used to prefer composition to inheritance, see en.wiki:Composition_over_inheritance.
    For more details on Mixins in Python, Fluent Python chapter 14 or Effective Python item 41 (links are free for MPG):
    - https://learning.oreilly.com/library/view/fluent-python-2nd/9781492056348/ch14.html#idm45517018812200
    - https://learning.oreilly.com/library/view/effective-python-90/9780134854717/ch05.xhtml#item41


    Arguments:
    - `name`: The name of the component.

    """
    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self.port: Set[Union[str, int]] = set()

    def _validate(self, dry_run):
        if not self.port:
            raise ValueError(f"{self} requires a mapping, None provided.")
        assert any(
            [c is not None for c in self.port]
        ), f"{self} has no mapped components. Please check the mapping."
        return super()._validate(dry_run)



""" A vessel, optionally with info on the chemical input contained. """
from typing import Optional
from ord_schema.proto.reaction_pb2 import ReactionInput

from flowchem.components.properties import Component


class Vessel(Component):
    """
    A generic vessel.

    Arguments:
    - `description`: The contents of the vessel.
    - `name`: The name of the vessel, if different from the description.

    Attributes:
    - `description`: The contents of the vessel.
    - `name`: The name of the vessel, if different from the description.
    """

    def __init__(self, description: Optional[str] = None, name: Optional[str] = None):
        super().__init__(name=name)
        self.description = description
        self.chemical = None

    def _validate(self, dry_run):
        # If chemical info are provided, they should be ReactionInput
        if self.chemical is not None:
            assert isinstance(
                self.chemical, ReactionInput
            ), "Vessel have a ReactionInput"



from typing import Optional

from loguru import logger
from flowchem.components.properties import TempControl


class DummyTempControl(TempControl):
    """
    A fake temp controller, used internally for testing.

    ::: danger
    This component can be used in a real protocol, although it doesn't actually exist.
    :::

    Arguments:
    - `port`: The available port names.
    - `name`: The name of the valve.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

    async def _update(self) -> None:
        logger.trace(f"Current temperature for {self.name} is {self.temp}")



from abc import ABC
from typing import Optional, Union, Dict, Any, Set

from flowchem.components.properties import (
    ActiveComponent,
    MultiportComponentMixin,
)


class Valve(MultiportComponentMixin, ActiveComponent, ABC):
    """
    A generic valve.

    Arguments:
    - `port`: The port numbers.
    - `name`: The name of the valve.
    """

    def __init__(
        self,
        port: Set[Union[int, str]],
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

        self.port = port
        # Base state is first port or 1 if no port is provided
        if port:
            self.setting = next(iter(port))
        else:
            self.setting = 1

        self._base_state: Dict[str, Any] = {"setting": 1}

    def _validate(self, dry_run):
        if not self.port:
            raise ValueError(f"The port names for valve {self} are not valid.")
        return super()._validate(dry_run)



""" Control module for the Vapourtec R4 heater """
import time

from flowchem.components.properties import ActiveComponent
from loguru import logger
from typing import Tuple, Optional

import aioserial

from flowchem.exceptions import DeviceError, InvalidConfiguration
from flowchem.units import flowchem_ureg

try:
    from flowchem.components.devices.Vapourtec.commands import (
        R4Command,
        VapourtecCommand,
    )

    HAS_VAPOURTEC_COMMANDS = True
except ImportError as e:
    HAS_VAPOURTEC_COMMANDS = False
    raise PermissionError(
        "Cannot redistribute Vapourtec commands... Contact Vapourtec to get them!"
    ) from e


class R4Heater(ActiveComponent):
    """R4 reactor heater control class."""

    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 19200,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }
    """ Virtual control of the Vapourtec R4 heating module. """

    def __init__(self, name: Optional[str] = None, **config):
        super().__init__(name)
        if not HAS_VAPOURTEC_COMMANDS:
            raise InvalidConfiguration(
                "R4Heater unusable: no Vapourtec Commands available.\n"
                "Contact your distributor to get the serial API documentation."
            )

        # Merge default settings, including serial, with provided ones.
        configuration = dict(R4Heater.DEFAULT_CONFIG, **config)
        try:
            self._serial = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as ex:
            raise InvalidConfiguration(
                f"Cannot connect to the R4Heater on the port <{config.get('port')}>"
            ) from ex

    async def _write(self, command: str):
        """Writes a command to the pump"""
        cmd = command + "\r\n"
        await self._serial.write_async(cmd.encode("ascii"))
        logger.debug(f"Sent command: {repr(command)}")

    async def _read_reply(self) -> str:
        """Reads the pump reply from serial communication"""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string.decode('ascii')}")
        return reply_string.decode("ascii")

    @staticmethod
    def parse_response(response: str) -> Tuple[bool, str]:
        """Split a received line in its components: success, reply"""
        try:
            if response[0:2] != "ER":
                return True, response.rstrip()
            else:
                return False, response.rstrip()
        except KeyError:
            raise DeviceError(f"Invalid reply {response}")

    async def write_and_read_reply(self, command: R4Command) -> str:
        """Main HamiltonPumpIO method.
        Sends a command to the pump, read the replies and returns it, optionally parsed"""
        self._serial.reset_input_buffer()
        await self._write(command.compile())
        response = await self._read_reply()

        if not response:
            raise InvalidConfiguration("No response received from heating module!")

        # Parse reply
        success, parsed_response = self.parse_response(response)

        return parsed_response

    async def wait_for_target_temp(self, channel: int):
        """Waits until the target channel has reached the desired temperature and is stable"""
        t_stable = False
        failure = 0
        while not t_stable:
            try:
                ret_code = await self.write_and_read_reply(
                    VapourtecCommand.TEMP.set_argument(channel)
                )
            except InvalidConfiguration as ex:
                ret_code = "N"
                failure += 1
                if failure > 3:
                    raise ex
            else:
                failure = 0

            if ret_code[:1] == "S":
                logger.debug(f"Target temperature reached on channel {channel}!")
                t_stable = True
            else:
                time.sleep(1)

    async def set_temperature(
        self, channel, target_temperature: str, wait: bool = False
    ):
        """Set temperature and optionally waits for S"""
        set_command = getattr(VapourtecCommand, f"SET_CH{channel}_TEMP")

        set_temperature = flowchem_ureg(target_temperature)
        # Float not accepted, must cast to int
        await self.write_and_read_reply(
            set_command.set_argument(round(set_temperature.m_as("°C")))
        )
        # Set temperature implies channel on
        await self.write_and_read_reply(VapourtecCommand.CH_ON.set_argument(channel))

        if wait:
            await self.wait_for_target_temp(channel)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/temperature/set", self.set_temperature, methods=["PUT"])

        return router


if __name__ == "__main__":
    import asyncio

    heat = R4Heater(port="COM11")

    async def main():
        """test function"""
        # noinspection PyArgumentEqualDefault
        await heat.set_temperature(0, "30 °C", wait=False)
        print("not waiting - default behaviour.")
        await heat.set_temperature(0, "30 °C", wait=True)
        print("actually I waited")

    asyncio.run(main())



from typing import Optional

from loguru import logger
from flowchem.components.properties import Valve


class DummyValve(Valve):
    """
    A fake valve, used internally for testing.

    ::: danger
    This component can be used in a real protocol, although it doesn't actually exist.
    :::

    Arguments:
    - `port`: The available port names.
    - `name`: The name of the valve.
    """

    def __init__(self, name: Optional[str] = None, port: set = None):
        if port is None:
            port = {"position_1", "position_2", "position_3"}
        super().__init__(name=name, port=port)

    async def _update(self) -> None:
        logger.trace(f"Switching {self.name} to port {self.setting}")



""" All devices should inherit from this class. """
from flowchem.components.properties import Component


class PassiveComponent(Component):
    """A non-connected, non-controllable, passive component."""

    _id_counter = 0



from datetime import timedelta

import pytest


from flowchem import DeviceGraph, Protocol
from flowchem.components.properties import Valve, Component
from flowchem.components.dummy import Dummy
from flowchem.components.stdlib import Pump, Vessel


@pytest.fixture
def device_graph():
    D = DeviceGraph()
    a, b = [Component() for _ in range(2)]
    pump = Pump("pump")
    D.add_device([a, b, pump])
    D.add_connection(a, b)
    D.add_connection(b, pump)
    return D


def test_create_protocol(device_graph):
    # test naming
    assert Protocol(device_graph, name="testing").name == "testing"
    assert Protocol(device_graph).name == "Protocol_0"
    assert Protocol(device_graph).name == "Protocol_1"


def test_add(device_graph):
    P = Protocol(device_graph)

    procedure = {
        "component": device_graph["pump"],
        "params": {"rate": "10 mL/min"},
        "start": 0,
        "stop": 300,
    }

    # test using duration
    P.add(device_graph["pump"], rate="10 mL/min", duration="5 min")
    assert P.procedures[0] == procedure

    # test adding with start and stop
    P.procedures = []
    P.add(device_graph["pump"], rate="10 mL/min", start="0 min", stop="5 min")
    assert P.procedures[0] == procedure

    # test adding with start and stop as timedeltas
    P.procedures = []
    P.add(
        device_graph["pump"], rate="10 mL/min", start=timedelta(seconds=0), stop=timedelta(minutes=5)
    )
    assert P.procedures[0] == procedure

    # test adding with duration as timedelta
    P.procedures = []
    P.add(device_graph["pump"], rate="10 mL/min", duration=timedelta(minutes=5))
    assert P.procedures[0] == procedure

    P = Protocol(device_graph)
    with pytest.raises(AssertionError):
        P.add(Pump("not in apparatus"), rate="10 mL/min", duration="5 min")

    # adding a class, not an instance of it
    with pytest.raises(ValueError):
        P.add(Pump, rate="10 mL/min", duration="5 min")

    # Not adding keyword args
    with pytest.raises(RuntimeError):
        P.add(device_graph["pump"], duration="5 min")

    # Invalid keyword for component
    with pytest.raises(ValueError):
        P.add(device_graph["pump"], active=False, duration="5 min")

    # Invalid dimensionality for kwarg
    with pytest.raises(ValueError):
        P.add(device_graph["pump"], rate="5 mL", duration="5 min")

    # No unit
    with pytest.raises(ValueError):
        P.add(device_graph["pump"], rate="5", duration="5 min")

    # Just the raw value without a unit
    with pytest.raises(ValueError):
        P.add(device_graph["pump"], rate=5, duration="5 min")

    # Providing stop and duration should raise error
    with pytest.raises(RuntimeError):
        P.add(device_graph["pump"], rate="5 mL/min", stop="5 min", duration="5 min")

    # stop time before start time
    with pytest.raises(ValueError):
        P.add([device_graph["pump"],  device_graph["pump"]], rate="10 mL/min", start="5 min", stop="4 min")


def test_add_dummy(device_graph):
    dummy = Dummy(name="dummy")
    device_graph.add_connection(dummy, "pump")
    P = Protocol(device_graph)
    with pytest.raises(ValueError):
        P.add(dummy, active=1)  # should be a bool!


# def test_add_valve(device_graph):
#     valve = Valve(port={1, 2})
#     bad_valve = Valve(port={1, 2})
#     pump1 = Pump("pump1")
#     pump2 = Pump("pump2")
#
#     device_graph.add_connection(valve, "pump", 2)
#     device_graph.add_connection(valve, pump1, 1)
#     device_graph.add_connection(bad_valve, "pump", 1)
#     device_graph.add_connection(bad_valve, pump2, 2)
#     P = Protocol(device_graph)
#
#     expected = [dict(start=0, stop=1, component=valve, params={"setting": 1})]
#
#     # directly pass the pump object
#     P.add(valve, setting=pump1, duration="1 sec")
#     assert P.procedures == expected
#     P.procedures = []
#
#     # using its name
#     P.add(valve, setting="pump1", duration="1 sec")
#     assert P.procedures == expected
#     P.procedures = []
#
#     # using its port number
#     P.add(valve, setting=1, duration="1 sec")
#     assert P.procedures == expected
#
#     with pytest.raises(ValueError):
#         P.add(valve, setting=3, duration="1 sec")
#
#     with pytest.raises(ValueError):
#         P.add(bad_valve, setting=3, duration="1 sec")


# def test_compile():
#     P = Protocol(A)
#     P.add([pump1, pump2], rate="10 mL/min", duration="5 min")
#     assert P._compile() == {
#         pump1: [
#             {"params": {"rate": "10 mL/min"}, "time": 0},
#             {"params": {"rate": "0 mL/min"}, "time": 300},
#         ],
#         pump2: [
#             {
#                 "params": {"rate": "10 mL/min"},
#                 "time": flowchem_ureg.parse_expression("0 seconds"),
#             },
#             {"params": {"rate": "0 mL/min"}, "time": 300},
#         ],
#     }
#
#
# def test_unused_component():
#     # raise warning if component not used
#     P = Protocol(A)
#     P.add(pump1, rate="10 mL/min", duration="5 min")
#     with pytest.warns(UserWarning, match="not used"):
#         P._compile()
#
#
# def test_switching_rates():
#     # check switching between rates
#     P = Protocol(A)
#     P.add([pump1, pump2], rate="10 mL/min", duration="5 min")
#     P.add(pump1, rate="5 mL/min", start="5 min", stop="10 min")
#     assert P._compile() == {
#         pump1: [
#             {"params": {"rate": "10 mL/min"}, "time": 0},
#             {"params": {"rate": "5 mL/min"}, "time": 300},
#             {"params": {"rate": "0 mL/min"}, "time": 600},
#         ],
#         pump2: [
#             {"params": {"rate": "10 mL/min"}, "time": 0},
#             {"params": {"rate": "0 mL/min"}, "time": 300},
#         ],
#     }
#
#
# def test_overlapping_procedures():
#     P = Protocol(A)
#     P.add(pump1, start="0 seconds", stop="5 seconds", rate="5 mL/min")
#     P.add(pump1, start="2 seconds", stop="5 seconds", rate="2 mL/min")
#     with pytest.raises(RuntimeError):
#         P._compile()
#
#
# def test_conflicting_continuous_procedures():
#     P = Protocol(A)
#     P.add(pump1, rate="5 mL/min", stop="1 sec")
#     P.add(pump1, rate="2 mL/min", stop="1 sec")
#     with pytest.raises(RuntimeError):
#         P._compile()
#
#
# def test_json():
#     P = Protocol(A)
#     P.add([pump1, pump2], rate="10 mL/min", duration="5 min")
#     assert json.loads(P.json()) == [
#         {
#             "start": 0,
#             "stop": 300,
#             "component": "pump1",
#             "params": {"rate": "10 mL/min"},
#         },
#         {
#             "start": 0,
#             "stop": 300,
#             "component": "pump2",
#             "params": {"rate": "10 mL/min"},
#         },
#     ]
#
#
# def test_yaml():
#     P = Protocol(A)
#     P.add([pump1, pump2], rate="10 mL/min", duration="5 min")
#     assert yaml.safe_load(P.yaml()) == json.loads(P.json())
